/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Index Model - IndexFieldAlias.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.model.index;

import java.util.ArrayList;
import java.util.List;

import javax.persistence.CollectionTable;
import javax.persistence.Column;
import javax.persistence.ElementCollection;
import javax.persistence.Entity;
import javax.persistence.JoinColumn;
import javax.persistence.Table;
import javax.persistence.UniqueConstraint;
import javax.validation.Valid;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import ch.unige.solidify.IndexConstants;
import ch.unige.solidify.SolidifyConstants;
import ch.unige.solidify.model.Label;
import ch.unige.solidify.rest.ResourceNormalized;

@Entity
@Table(uniqueConstraints = { @UniqueConstraint(columnNames = { "indexName", "alias" }) })
public class IndexFieldAlias extends ResourceNormalized {

  /**
   * The name of the ElasticSearch index the field is for
   */
  @NotNull
  @Size(min = 1)
  private String indexName;

  /**
   * The name that will be exposed to clients (as a field or a facet)
   */
  @NotNull
  @Size(min = 1)
  private String alias;

  /**
   * The field path in the index
   */
  @NotNull
  @Size(min = 1)
  private String field;

  /**
   * Should a facet be created for this field ? default to false
   */
  @NotNull
  private Boolean facet;

  /**
   * Is used by system
   */
  @NotNull
  private Boolean system;

  /**
   * Minimum number of occurrences of a facet value to be returned in facet results if the
   * IndexFieldAlias is a facet
   */
  private Integer facetMinCount;

  /**
   * Maximum number of facets values to return if the IndexFieldAlias is a facet
   */
  private Integer facetLimit;

  /**
   * Determines the position of a facet in the user interface
   */
  private Integer facetOrder;

  /**
   * Number of facet values displayed by default in the user interface
   */
  private Integer facetDefaultVisibleValues;

  @Valid
  @ElementCollection
  @Column(name = IndexConstants.DB_INDEX_FIELD_ALIAS_ID)
  @CollectionTable(joinColumns = {
          @JoinColumn(name = IndexConstants.DB_INDEX_FIELD_ALIAS_ID) }, uniqueConstraints = @UniqueConstraint(columnNames = {
                  IndexConstants.DB_INDEX_FIELD_ALIAS_ID, SolidifyConstants.DB_LANGUAGE_ID }))
  private List<Label> labels = new ArrayList<>();

  public String getIndexName() {
    return this.indexName;
  }

  public void setIndexName(String indexName) {
    this.indexName = indexName;
  }

  public String getAlias() {
    return this.alias;
  }

  public void setAlias(String alias) {
    this.alias = alias;
  }

  public String getField() {
    return this.field;
  }

  public void setField(String field) {
    this.field = field;
  }

  public Boolean isFacet() {
    return this.facet;
  }

  public void setFacet(Boolean facet) {
    this.facet = facet;
  }

  public Boolean isSystem() {
    return system;
  }

  public void setSystem(Boolean system) {
    this.system = system;
  }

  public Integer getFacetMinCount() {
    return this.facetMinCount;
  }

  public void setFacetMinCount(Integer facetMinCount) {
    this.facetMinCount = facetMinCount;
  }

  public Integer getFacetLimit() {
    return this.facetLimit;
  }

  public void setFacetLimit(Integer facetLimit) {
    this.facetLimit = facetLimit;
  }

  public Integer getFacetOrder() {
    return this.facetOrder;
  }

  public void setFacetOrder(Integer facetOrder) {
    this.facetOrder = facetOrder;
  }

  public Integer getFacetDefaultVisibleValues() {
    return this.facetDefaultVisibleValues;
  }

  public void setFacetDefaultVisibleValues(Integer facetDefaultVisibleValues) {
    this.facetDefaultVisibleValues = facetDefaultVisibleValues;
  }

  public List<Label> getLabels() {
    return this.labels;
  }

  public void setLabels(List<Label> labels) {
    this.labels = labels;
  }

  @Override
  public void init() {
    if (this.isFacet() == null) {
      this.setFacet(false);
    }
    if (this.isSystem() == null) {
      this.setSystem(false);
    }
  }

  @Override
  public String managedBy() {
    return IndexConstants.INDEX_MODULE;
  }

  @Override
  public boolean isDeletable() {
    return !this.isSystem();
  }
}
